from .demo_func import DemoFunc
from .demo_func_ackley import DemoFuncAckley
from .demo_func_brown import DemoFuncBrown
from .demo_func_grienwank import DemoFuncGrienwank
from .demo_func_michalewicz import DemoFuncMichalewicz
from .demo_func_piston import DemoFuncPiston
from .demo_func_rastrigin import DemoFuncRastrigin
from .demo_func_rosenbrock import DemoFuncRosenbrock
from .demo_func_schaffer import DemoFuncSchaffer
from .demo_func_schwefel import DemoFuncSchwefel
